﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text.Json;
using System.Xml;

namespace ERNW.PowerShell.AdministrativeTemplate.AdmxTemplate
{
    public class TemplateParser
    {
        public TemplateParser(string admxTemplateFilePath, string language)
        {
            Initialize(admxTemplateFilePath, language);
        }

        public List<TemplateObject> TemplateObject { get; set; }

        internal void Initialize(string admxTemplateFilePath, string language)
        {
            XmlDocument xAdmlTemplate = new XmlDocument();
            xAdmlTemplate.XmlResolver = null;
            XmlDocument xAdmxTemplate = new XmlDocument();
            xAdmxTemplate.XmlResolver = null;
            TemplateObject = new List<TemplateObject>();


            FileInfo admxTemplateFileInfo = new FileInfo(admxTemplateFilePath);
            if (admxTemplateFileInfo.Exists)
            {

                FileInfo admlTemplateFileInfo = new FileInfo(Path.Combine(admxTemplateFileInfo.DirectoryName, language, admxTemplateFileInfo.Name.Replace(".admx", ".adml")));
                if (admlTemplateFileInfo.Exists)
                {
                    xAdmlTemplate.Load(admlTemplateFileInfo.FullName);
                }

                xAdmxTemplate.Load(admxTemplateFileInfo.FullName);
                XmlNamespaceManager ns = new XmlNamespaceManager(xAdmxTemplate.NameTable);
                ns.AddNamespace("pd", xAdmxTemplate.DocumentElement.NamespaceURI);

                foreach (XmlNode xPol in xAdmxTemplate.SelectNodes("/pd:policyDefinitions/pd:policies/pd:policy", ns))
                {
                    TemplateObject oRes = new TemplateObject();
                    oRes.SettingAttributes.elements = new ArrayList();
                    //elements = new List<Element>();
                    oRes.GeneralAttributes.FileName = admxTemplateFileInfo.FullName;
                    oRes.GeneralAttributes.displayName = sResourceStringLookup(xPol.Attributes["displayName"].InnerText.Replace("$(string.", "").TrimEnd(')'), xAdmlTemplate);
                    oRes.GeneralAttributes.name = xPol.Attributes["name"].InnerText;
                    oRes.GeneralAttributes.state = PolicyState.NotConfigured;

                    if (string.IsNullOrEmpty(oRes.GeneralAttributes.displayName))
                        oRes.GeneralAttributes.displayName = oRes.GeneralAttributes.name;

                    if (xPol.Attributes["explainText"] != null)
                        oRes.explainText = sResourceStringLookup(xPol.Attributes["explainText"].InnerText.Replace("$(string.", "").TrimEnd(')'), xAdmlTemplate);

                    oRes.SettingAttributes.key = xPol.Attributes["key"].InnerText;
                    if (xPol.Attributes["presentation"] != null)
                        oRes.presentation = sPresentationStringLookup(xPol.Attributes["presentation"].InnerText.Replace("$(presentation.", "").TrimEnd(')'), xAdmlTemplate);
                    switch (xPol.Attributes["class"].InnerText)
                    {
                        case "Machine":
                            oRes.GeneralAttributes.classType = ClassType.Machine;
                            break;
                        case "User":
                            oRes.GeneralAttributes.classType = ClassType.User;
                            break;
                    }



                    if (xPol.Attributes["valueName"] != null)
                    {
                        if (xPol["enabledValue"] != null)
                        {
                            PolEnableElement oElem = new PolEnableElement();
                            oElem.ValueName = xPol.Attributes["valueName"].InnerText;
                            oElem.ValueType = ValueType.PolicyEnable;

                            if (xPol["enabledValue"]["decimal"] != null)
                            {
                                try
                                {
                                    oElem.enabledValue = uint.Parse(xPol["enabledValue"]["decimal"].Attributes["value"].Value);
                                }
                                catch (Exception ex)
                                {
                                    Debug.WriteLine(ex.Message);
                                }
                            }

                            if (xPol["disabledValue"] != null)
                            {
                                if (xPol["disabledValue"]["decimal"] != null)
                                {
                                    try
                                    {
                                        oElem.disabledValue = uint.Parse(xPol["disabledValue"]["decimal"].Attributes["value"].Value);
                                    }
                                    catch (Exception ex)
                                    {
                                        Debug.WriteLine(ex.Message);
                                    }
                                }
                            }

                            oRes.SettingAttributes.elements.Add(oElem);
                        }
                        else
                        {
                            PolEnableElement oElem = new PolEnableElement();
                            oElem.ValueName = xPol.Attributes["valueName"].InnerText;
                            oElem.ValueType = ValueType.PolicyEnable;
                            oElem.enabledValue = 1;
                            oElem.disabledValue = 0;

                            oRes.SettingAttributes.elements.Add(oElem);
                        }
                    }

                    if (xPol["elements"] != null)
                    {
                        foreach (XmlNode xElem in xPol["elements"].ChildNodes)
                        {
                            if (xElem.Name == "#comment")
                                continue;
                            if (xElem.Name == "#text")
                                continue;

                            Element oElem = new Element();

                            switch (xElem.Name.ToLower())
                            {
                                case "decimal":
                                    oElem = new decimalElement();
                                    oElem.ValueType = ValueType.Decimal;
                                    if (xElem.Attributes["minValue"] != null)
                                    {
                                        if (uint.TryParse(xElem.Attributes["minValue"].InnerText, out uint iRes))
                                        {
                                            ((decimalElement)oElem).minValue = iRes;
                                        }
                                    }
                                    if (xElem.Attributes["maxValue"] != null)
                                    {
                                        if (uint.TryParse(xElem.Attributes["maxValue"].InnerText, out uint iRes))
                                        {
                                            ((decimalElement)oElem).maxValue = iRes;
                                        }
                                    }
                                    break;
                                case "enum":
                                    oElem = new enumElement();
                                    oElem.ValueType = ValueType.Enum;
                                    ((enumElement)oElem).valueList = new Dictionary<string, string>();
                                    foreach (XmlNode xElemItem in xElem.SelectNodes("pd:item", ns))
                                    {
                                        try
                                        {
                                            string sDisplayName = sResourceStringLookup(xElemItem.Attributes["displayName"].Value.Replace("$(string.", "").TrimEnd(')'), xAdmlTemplate);
                                            if (string.IsNullOrEmpty(sDisplayName))
                                                sDisplayName = xElemItem.Attributes["displayName"].Value.Replace("$(string.", "").TrimEnd(')');
                                            if (xElemItem["value"].FirstChild.Name == "decimal")
                                            {
                                                if (xElemItem["value"].FirstChild.Attributes.Count > 0)
                                                {
                                                    ((enumElement)oElem).valueList.Add(sDisplayName, xElemItem["value"].FirstChild.Attributes["value"].Value);
                                                }
                                                else
                                                {
                                                    ((enumElement)oElem).valueList.Add(sDisplayName, xElemItem["value"].FirstChild.InnerText);
                                                }
                                            }
                                            if (xElemItem["value"].FirstChild.Name == "string")
                                                ((enumElement)oElem).valueList.Add(sDisplayName, xElemItem["value"].FirstChild.InnerText);
                                            string sDefault = sPresentationdefaultValue(xElem.Attributes["id"].Value, oRes.presentation);
                                            if (!string.IsNullOrEmpty(sDefault))
                                            {
                                                int.TryParse(sDefault, out int iDef);

                                                ((enumElement)oElem).defaultItem = iDef;
                                            }
                                        }
                                        catch (Exception ex)
                                        {
                                            Debug.WriteLine(ex.Message);
                                        }
                                    }
                                    break;
                                case "boolean":
                                    oElem = new decimalElement();
                                    oElem.ValueType = ValueType.Boolean;
                                    ((decimalElement)oElem).minValue = 0;
                                    ((decimalElement)oElem).maxValue = 1;
                                    oElem.value = sPresentationdefaultValue(xElem.Attributes["id"].Value, oRes.presentation);
                                    break;
                                case "list":
                                    oElem = new ListElement();
                                    oElem.ValueType = ValueType.List;
                                    ((ListElement)oElem).valueList = new Dictionary<string, string>();
                                    if (xElem.Attributes["additive"] != null)
                                    {
                                        if (bool.TryParse(xElem.Attributes["additive"].Value, out bool bRes))
                                        {
                                            ((ListElement)oElem).additive = bRes;
                                        }
                                        else
                                            ((ListElement)oElem).additive = null;
                                    }
                                    if (xElem.Attributes["explicitValue"] != null)
                                    {
                                        if (bool.TryParse(xElem.Attributes["explicitValue"].Value, out bool bRes))
                                        {
                                            ((ListElement)oElem).explicitValue = bRes;
                                        }
                                        else
                                            ((ListElement)oElem).explicitValue = null;
                                    }
                                    break;
                                case "text":
                                    oElem = new TextElement();
                                    oElem.ValueType = ValueType.Text;
                                    break;
                                default:
                                    xElem.Name.ToString();
                                    break;
                            }

                            oElem.value = sPresentationdefaultValue(xElem.Attributes["id"].InnerText, oRes.presentation);

                            //List do not have a Value, they have List of Valuenames and Values                            
                            if (oElem.ValueType != ValueType.List)
                                oElem.ValueName = xElem.Attributes["valueName"].InnerText;

                            if (xElem.Attributes["required"] != null)
                            {
                                if (bool.TryParse(xElem.Attributes["required"].InnerText, out bool bReq))
                                    oElem.required = bReq;
                            }

                            if (xElem.Attributes["key"] != null)
                            {
                                oElem.key = xElem.Attributes["key"].Value;
                            }

                            oRes.SettingAttributes.elements.Add(oElem);
                        }
                    }

                    if (xPol["enabledList"] != null)
                    {
                        EnableListElement oElem = new EnableListElement();

                        oElem.ValueType = ValueType.Enum;
                        oElem.enabledValueList = new List<EnabledList>();

                        foreach (XmlNode xElem in xPol["enabledList"].SelectNodes("pd:item", ns))
                        {
                            try
                            {
                                EnabledList oResList = new EnabledList();

                                if (xElem["value"].FirstChild.Name == "decimal")
                                {
                                    oResList.type = ValueType.Decimal;
                                    oResList.key = xElem.Attributes["key"].Value;
                                    oResList.valueName = xElem.Attributes["valueName"].Value;
                                    oResList.value = xElem["value"].FirstChild.Attributes["value"].Value;
                                }
                                if (xElem["value"].FirstChild.Name == "string")
                                {
                                    oResList.type = ValueType.Text;
                                    oResList.key = xElem.Attributes["key"].Value;
                                    oResList.valueName = xElem.Attributes["valueName"].Value;
                                    if (xElem["value"].FirstChild.Attributes.Count > 0)
                                    {
                                        oResList.value = xElem["value"].FirstChild.Attributes["value"].Value;
                                    }
                                    else
                                    {
                                        oResList.value = xElem["value"].FirstChild.InnerText;
                                    }
                                }

                                oElem.enabledValueList.Add(oResList);
                            }
                            catch (Exception ex)
                            {
                                Debug.WriteLine(ex.Message);
                            }
                        }
                        oRes.SettingAttributes.elements.Add(oElem);
                    }

                    if (xPol["disabledList"] != null)
                    {
                        EnableListElement oElem = new EnableListElement();

                        oElem.ValueType = ValueType.Enum;
                        oElem.disabledValueList = new List<EnabledList>();

                        foreach (XmlNode xElem in xPol["disabledList"].SelectNodes("pd:item", ns))
                        {
                            try
                            {
                                EnabledList oResList = new EnabledList();

                                if (xElem["value"].FirstChild.Name == "decimal")
                                {
                                    oResList.type = ValueType.Decimal;
                                    oResList.key = xElem.Attributes["key"].Value;
                                    oResList.valueName = xElem.Attributes["valueName"].Value;
                                    oResList.value = xElem["value"].FirstChild.Attributes["value"].Value;
                                }
                                if (xElem["value"].FirstChild.Name == "string")
                                {
                                    oResList.type = ValueType.Text;
                                    oResList.key = xElem.Attributes["key"].Value;
                                    oResList.valueName = xElem.Attributes["valueName"].Value;
                                    if (xElem["value"].FirstChild.Attributes.Count > 0)
                                    {
                                        oResList.value = xElem["value"].FirstChild.Attributes["value"].Value;
                                    }
                                    else
                                    {
                                        oResList.value = xElem["value"].FirstChild.InnerText;
                                    }
                                }

                                oElem.disabledValueList.Add(oResList);
                            }
                            catch (Exception ex)
                            {
                                Debug.WriteLine(ex.Message);
                            }


                        }


                        oRes.SettingAttributes.elements.Add(oElem);
                    }

                    TemplateObject.Add(oRes);

                }
            }
        }

        public string TemplateObjectJsonSerialize<TValue>(TValue value)
        {
            return JsonSerializer.Serialize<TValue>(value).ToString();
        }

        internal string sResourceStringLookup(string ID, XmlDocument xAdmlTemplate)
        {
            string sResult = "";
            try
            {
                if (!string.IsNullOrEmpty(xAdmlTemplate.OuterXml))
                {
                    if (string.IsNullOrEmpty(xAdmlTemplate.DocumentElement.NamespaceURI))
                    {
                        var oRes = xAdmlTemplate.SelectSingleNode("/policyDefinitionResources/resources/stringTable/string[@id=\"" + ID + "\"]");
                        if (oRes != null)
                            return oRes.InnerText;
                        else
                            return ID;
                    }
                    else
                    {
                        XmlNamespaceManager ns = new XmlNamespaceManager(xAdmlTemplate.NameTable);
                        //ns.AddNamespace("pd", "http://schemas.microsoft.com/GroupPolicy/2006/07/PolicyDefinitions");
                        ns.AddNamespace("pd", xAdmlTemplate.DocumentElement.NamespaceURI);


                        var oRes = xAdmlTemplate.SelectSingleNode("/pd:policyDefinitionResources/pd:resources/pd:stringTable/pd:string[@id=\"" + ID + "\"]", ns);

                        if (oRes != null)
                        {
                            if (string.IsNullOrEmpty(oRes.InnerText))
                            {
                                oRes.InnerText.ToString();
                                return ID;
                            }
                            return oRes.InnerText;
                        }
                        else
                            return ID;
                    }

                }
            }
            catch (Exception ex)
            {
                ex.Message.ToString();
            }

            return sResult;
        }
        internal string sPresentationStringLookup(string ID, XmlDocument xAdmlTemplate)
        {
            string sResult = "";
            try
            {
                if (!string.IsNullOrEmpty(xAdmlTemplate.OuterXml))
                {
                    XmlNamespaceManager ns = new XmlNamespaceManager(xAdmlTemplate.NameTable);
                    ns.AddNamespace("pd", "http://schemas.microsoft.com/GroupPolicy/2006/07/PolicyDefinitions");
                    var oRes = xAdmlTemplate.SelectSingleNode("/pd:policyDefinitionResources/pd:resources/pd:presentationTable/pd:presentation[@id=\"" + ID + "\"]", ns);

                    if (oRes != null)
                        return oRes.InnerXml;
                    else return "";
                }
            }
            catch { }

            return sResult;
        }

        internal string sPresentationdefaultValue(string refId, string presentation)
        {
            string sResult = "";
            try
            {
                if (!string.IsNullOrEmpty(presentation))
                {
                    XmlDocument xPres = new XmlDocument();
                    xPres.LoadXml("<root>" + presentation + "</root>");
                    XmlNamespaceManager ns = new XmlNamespaceManager(xPres.NameTable);
                    ns.AddNamespace("pd", "http://schemas.microsoft.com/GroupPolicy/2006/07/PolicyDefinitions");
                    var vDefVal = xPres.SelectSingleNode("/root/*[contains(@refId,'" + refId + "')]", ns);
                    if (vDefVal != null)
                    {
                        if (vDefVal.Attributes["defaultValue"] != null)
                            sResult = vDefVal.Attributes["defaultValue"].InnerText;
                        if (vDefVal.Attributes["defaultChecked"] != null)
                        {
                            if (vDefVal.Attributes["defaultChecked"].InnerText == "true")
                                sResult = "1";
                            else
                                sResult = "0";
                        }
                        if (vDefVal.Attributes["defaultItem"] != null)
                            sResult = vDefVal.Attributes["defaultItem"].InnerText;
                    }
                }
            }
            catch (Exception ex)
            {
                ex.Message.ToString();
            }

            return sResult;
        }
    }
}
