﻿using Microsoft.Data.Sqlite;
using System;
using System.Runtime.InteropServices;

namespace ERNW.PowerShell.AdministrativeTemplate.AdmxSqliteProcessor
{
    public class SqliteProcessor : IDisposable
    {
        private bool isDisposed;
        private IntPtr nativeResource = Marshal.AllocHGlobal(100);

        // Dispose() calls Dispose(true)
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        // The bulk of the clean-up code is implemented in Dispose(bool)
        protected virtual void Dispose(bool disposing)
        {
            if (isDisposed) return;

            if (disposing)
            {
                // free managed resources
                Connection.Dispose();

            }

            // free native resources if there are any.
            if (nativeResource != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(nativeResource);
                nativeResource = IntPtr.Zero;
            }

            isDisposed = true;
        }

        public SqliteProcessor()
        {
            Dispose(false);
        }

        public string BaseIDPolicyTable { get; set; }
        public string BaseID { get; set; }
        public string BaseTime { get; set; } = DateTime.Now.ToString("ddMMyyyy:HHmmss");
        public SqliteConnection Connection { get; private set; }
        public void OpenAdmxSqliteDb(string baseConnectionString)
        {
            if (System.Data.ConnectionState.Closed == 0)
            {
                string connectionString = new SqliteConnectionStringBuilder()
                {
                    Mode = SqliteOpenMode.ReadWriteCreate,
                    DataSource = baseConnectionString

                }.ToString();

                Connection = new SqliteConnection(connectionString);

                Connection.Open();

            }
        }

        public void ExecuteAdmxNonQueryAsync(string query)
        {
            using (var command = Connection.CreateCommand())
            {
                command.CommandText = query;
                command.ExecuteNonQueryAsync();
            }
        }

        public void ExecuteAdmxNonQuery(string query)
        {
            using (var command = Connection.CreateCommand())
            {
                command.CommandText = query;
                command.ExecuteNonQuery();
            }
        }


        public void CreateSqliteAdmxTable(string baseConnectionString)
        {
            long value = long.Parse(DateTime.Now.ToString("ddMMyyyyHHmmss"));
            string Chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
            //Encode String to Base36
            while (value > 0)
            {
                BaseID = Chars[(int)(value % 36)] + BaseID;
                value /= 36;
            }

            OpenAdmxSqliteDb(baseConnectionString);
            ExecuteAdmxNonQuery(@"CREATE TABLE IF NOT EXISTS BaseTable([base_id] TEXT NOT NULL, [base_name] TEXT NOT NULL, [base_time] TEXT NOT NULL, [base_osversion] TEXT NOT NULL, [base_count] INTEGER NOT NULL, PRIMARY KEY(base_id))");

            BaseIDPolicyTable = BaseID + "_BasePolicyTable";
            ExecuteAdmxNonQuery($@"CREATE TABLE IF NOT EXISTS '{BaseIDPolicyTable}'([base_id] TEXT NOT NULL, [policies] TEXT NOT NULL,                                            
                                            FOREIGN KEY (base_id)
                                                REFERENCES BaseTable (base_id)
                                                    ON DELETE CASCADE
                                         )");
        }
    }
}
